#!/usr/bin/env python

import math
from fftwlib import charMap

def rebuildSpellMes(fin):
    print("  Rebuilding:")
    lines,textLines=[],[]
    charMap.fixCharacterSet()
    with open(fin+"-unt",encoding="utf8",errors="ignore") as file:
        for line_number,line in enumerate(file,start=1):
            lines.append(line[:-1])
    with open(fin+"-utt.txt",encoding="utf8",errors="ignore") as file:
        for line_number,line in enumerate(file,start=1):
            textLines.append(line[:-1])
    with open(fin+"-utt","w",encoding="utf8",errors="ignore",newline="") as file:
        sline=""
        for line in lines:
            if line=="UNT":
                sline="{Color 08}{Caster}{Color 00}{Newline}"
                line=textLines.pop(0)
                if not line:
                    sline=""
                else:
                    if line[0:8]=="{format}":
                        sline+=line[8:]
                    else:
                        if line.count("!")==2:
                            line=line.split("! ",1)
                            line[0]=line[0]+"!"
                        else:
                            line=line.rsplit(". ",1)
                            line[0]=line[0]+"."
                        if getStringWidth(line[0])>160:
                            line[0]=setLineWidth(line[0],0,150)
                            sline+=line[0][:-2]+"{Delay 2A}"+line[0][-2]\
                                   +"{Delay 00}"+"{LS}"+line[1][:-1]\
                                   +"{Delay 24}"+line[1][-1]
                        else:
                            sline+=line[0][:-1]+"{Delay 2A}"+line[0][-1]\
                                   +"{Delay 00}"+"{Newline}"+line[1][:-1]\
                                   +"{Delay 24}"+line[1][-1]
            elif "Close}" in line:
                sline+=line+"\n"
                file.writelines(sline)
                sline=""
            elif "{zeroes(" in line:
                if sline:
                    file.writelines(sline+"\n")
                file.writelines(line+"\n")
    print("   Done.")
    return
def rebuildSNPLMes(fin,NOVELS=6,NNOV=5):
    charMap.fixCharacterSet()
    print("  Rebuilding:")
    for i in range(1,NOVELS+1):
        if i==NNOV:
            continue
        with open(fin+"-0"+str(i)+"-utt","w",encoding="utf8",errors="ignore",newline="") as outFile:
            with open(fin+"-0"+str(i)+"-utt.txt",encoding="utf8",errors="ignore") as txtFile:
                txtFile.readline()
                prevWidth,emptyEntry=0,True
                with open(fin+"-0"+str(i)+"-unt",encoding="utf8",errors="ignore") as file:
                    deleteNL=False
                    for line_number,line in enumerate(file,start=1):
                        if "UNT" in line:
                            emptyEntry,text=False,txtFile.readline()
                            if text!="{Delete}\n":
                                line=reapplyDelaysAndSound(line.split("UNT")[0],setLineWidth(text,prevWidth))
                                outFile.writelines(line)
                                if not checkIfTextEndsWNL(line):
                                    prevWidth=getStringWidth(line.split("{Newline}")[-1])
                                else:
                                    prevWidth=0
                            else:
                                deleteNL=True
                        else:
                            if deleteNL:
                                deleteNL=False
                            else:
                                if "Close}" in line:
                                    if not emptyEntry:
                                        line="{LS}"+line
                                    emptyEntry=True
                                else:
                                    emptyEntry=False
                                outFile.writelines(line)
                                if ("{Newline}" in line) or ("Close}" in line):
                                    prevWidth=0
    for i in range(1,NOVELS+1):
        lines=[]
        if i==NNOV:
            continue
        print("   Fixing newlines in: "+fin+"-0"+str(i))
        with open(fin+"-0"+str(i)+"-utt",encoding="utf8",errors="ignore") as file:
            for line_number,line in enumerate(file,start=1):
                lines.append(line)
        fixNewLines(lines,fin+"-0"+str(i))
    lines=[]
    for i in range(1,NOVELS+1):
        nfile=fin+"-0"+str(i)+"-utt"
        if i==NNOV:
            nfile=fin+"-0"+str(i)+"-unt"
        with open(nfile,encoding="utf8",errors="ignore") as file:
            for line_number,line in enumerate(file,start=1):
                lines.append(line)
    with open(fin+"-utt","w",encoding="utf8",errors="ignore",newline="") as file:
        for line in lines:
            file.writelines(line)
    print("   Done.")
    return
def extractEntriesNovels(fin,fou,NNOV=5):
    lines,novels=[],[]
    with open(fin,encoding="utf8",errors="ignore") as file:
        for line_number,line in enumerate(file,start=1):
            lines.append(line)
            if "{zeroes(" in line:
                novels.append(lines)
                lines=[]
    if lines:
        novels.append(lines)
    for idx in range(0,len(novels)):
        decodeLines(novels[idx],fou+"-0"+str(idx+1),idx+1!=NNOV)
    return
def decodeLines(lines,novel,touch):
    with open(novel+"-unt.txt","w",encoding="utf8",errors="ignore",newline="") as file:
        file.writelines("#Start\n")
    padding=""
    if "{zeroes(" in lines[-1]:
        padding=lines[-1]
    entries=rewriteEntries(getEntries(lines),touch)
    with open(novel+"-unt","w",encoding="utf8",errors="ignore",newline="") as file:
        for entry in entries:
            for line in entry:
                if getText(line) and touch:
                    file.writelines(translateT(line,novel))
                else:
                    file.writelines(line)
        if padding:
            file.writelines(padding)
    return
def fixNewLines(lines,novel):
    padding=""
    if "{zeroes(" in lines[-1]:
        padding=lines[-1]
    entries=getEntries(lines)
    with open(novel+"-utt","w",encoding="utf8",errors="ignore",newline="") as file:
        num_entry=0
        for entry in entries:
            num_entry+=1
            if hasEntryMT9NL(entry):
                entry=fixEntryTo9NL(entry,num_entry)
            for line in entry:
                file.writelines(line)
        if padding:
            file.writelines(padding)
    return
def hasEntryMT9NL(entry):
    countNL=0
    for line in entry:
        countNL+=line.count("{Newline}")
    return countNL>9
def fixEntryTo9NL(entry,num):
    countNL=0
    for line in entry:
        countNL+=line.count("{Newline}")
    countEL,lastLineNL=0,True
    for line in entry:
        if not lastLineNL:
            if checkIfTextEndsWNL(line):
                lastLineNL=True
        elif lineOnlyHasDorNL(line):
            countEL+=1
        else:
            lastLineNL=False
    startEL=countEL-(countNL-9)
    if startEL<0:
        print("    Can't fix entry: "+str(num)+".")
        return entry
    else:
        fixedEntry,countCEL,lastLineNL=[],0,True
        for line in entry:
            if not lastLineNL:
                fixedEntry.append(line)
                if checkIfTextEndsWNL(line):
                    lastLineNL=True
            elif lineOnlyHasDorNL(line):
                if countCEL<startEL:
                    fixedEntry.append(line)
                countCEL+=1
            else:
                lastLineNL=False
                fixedEntry.append(line)
        return fixedEntry
def lineOnlyHasDorNL(line):
    chars=charSplitter(line)
    for char in chars:
        if char!=" " and char!="{LS}" and char!="{SP}"\
                and (not "{Delay" in char)\
                and char!="{Newline}"\
                and (not "{Pad0" in char):
            return False
    return True
def translateT(text,novel):
   with open(novel+"-unt.txt",'a+',encoding="utf8",errors="ignore",newline="") as file:
       file.writelines(cleanInst(text))
   return text[0:text.rfind('}')+1]+"UNT\n"
def getEntries(lines):
    entries,lines2=[],[]
    for line in lines:
        lines2.append(line)
        if "Close}" in line:
            entries.append(lines2)
            lines2=[]
    return entries
def rewriteEntries(entries,touch):
    fixedEntries=[]
    for entry in entries:
        ending=entry.pop(-1)
        if entry and touch:
            entry=processEntry(entry)
        entry.append(ending)
        fixedEntries.append(entry)
    return fixedEntries
def extractEntriesSQ(fin,fou):
    lines,sline=[],""
    with open(fin,encoding="utf8",errors="ignore") as file:
        for line_number,line in enumerate(file,start=1):
            line=line[:-1]
            if "Close}" in line:
                if sline:
                    lines.append(sline+"\n")
                lines.append(line+"\n")
                sline=""
            elif "{zeroes(" in line:
                if sline:
                    lines.append(sline+"\n")
                lines.append(line+"\n")
                sline=""
            else:
                sline+=line
    with open(fou+"-unt","w",encoding="utf8",errors="ignore",newline="") as file:
        for line in lines:
            if "Close}" in line or "{zeroes(" in line:
                file.writelines(line)
            else:
                file.writelines("UNT\n")
    with open(fou+"-unt.txt","w",encoding="utf8",errors="ignore",newline="") as file:
        for line in lines:
            if "Close}" in line or "{zeroes(" in line:
                continue
            else:
                file.writelines(cleanInst(line))
    return
def charSplitterAndCompress(line):
    lparts,newLine,compress=charSplitter(line),"",""
    for char in lparts:
        if not "{" in char:
            compress+=char
        else:
            if compress.strip():
                compress=charMap.charEngCompress(compress)
            newLine+=compress+char
            compress=""
    if compress.strip():
        compress=charMap.charEngCompress(compress)
    return charSplitter(newLine+compress+"\n")
def charSplitter(line):
    if line[-1]!="\n":
        line+="\n"
    list=[]
    while len(line)>1:
        off1=line.find("{")
        list.extend([char for char in line[:off1]])
        if off1!=-1:
            off2=line.find("}")
            if off2==-1:
                return list
            inst=line[off1:off2+1]
            if "{Pad0"!=inst[0:5]:
                list.append(inst)
            line=line[off2+1:]
        else:
            break
    return list
def cleanInst(line):
    inst,sline,skip="","",0
    for i in line:
        if i=="{":
            inst,skip=i,skip+1
        elif i=="}" and skip>0:
            inst,skip=inst+i,skip-1
            if "{SP}"==inst or "{LS}"==inst:
                sline+=" "
        elif skip==0:
            sline+=i
        else:
            inst+=i
    return sline
def getStringWidth(line):
    width,line=0,charSplitter(line)
    for char in line:
        width+=charMap.getCharWitdh(char)
    return width
def setLineWidth(line,prevWidth,MAXWIDTH=200):
    lines,processedLines=line.split("{Newline}"),[]
    for sline in lines:
        words=sline.split()
        currentLine=getInitialSpaces(sline)
        if words:
            word=words.pop(0)
            if getStringWidth(word)==0:
                currentLine+=word
            elif (getStringWidth(word)+prevWidth)<=MAXWIDTH:
                currentLine+=word
            else:
                processedLines.append("")
                prevWidth=0
                currentLine+=word
        for word in words:
            if (getStringWidth(currentLine+" "+word)+prevWidth)<=MAXWIDTH:
                currentLine+=" "+word
            else:
                processedLines.append(currentLine)
                prevWidth=0
                currentLine=word
        if currentLine:
            processedLines.append(currentLine)
    processedLine=""
    for line in processedLines:
        processedLine+=line+"{Newline}"
    line=processedLine[0:processedLine.rfind('}')-8]
    if line:
        line+="\n"
    return line
def getInitialSpaces(line):
    initialSpaces=""
    for char in line:
        if char==" ":
            initialSpaces+=" "
        else:
            break
    return initialSpaces
def processEntry(entry):
    fixedEntry,processedText,processedInst=[],"",""
    lastDelay,typingSound,notSpreadTS=1,0,0
    while entry:
        currentLine=entry.pop(0)
        while currentLine:
            extraText=getText(currentLine)
            if typingSound==0 and extraText:
                typingSound=isTypingSound(currentLine,currentLine,2)
                if isTypingSound(currentLine,currentLine,0)>0:
                    extraText=getText(currentLine,2)
                    notSpreadTS=2
            processedText+=extraText
            if processedText:
                if processedText[-1]=="\n":
                    processedText=processedText[:-1]
            currentLine=removeString(currentLine,extraText)
            if extraText:
                if processedInst:
                    lastDelay=getLastDelay(processedInst)
                    if processedInst[-1]!="\n":
                        processedInst+="\n"
                    fixedEntry.append(processedInst)
                    processedInst=""
            else:
                if processedText:
                    avgDelay=getAverageDelay(processedText,lastDelay)
                    addNewline=checkIfTextEndsWNL(processedText)
                    processedText=cleanInst(processedText)
                    if typingSound==1:
                        processedText="{0xF55C}{0xF63A}"+processedText
                    elif typingSound==2:
                        processedText="{0xF55C}{0xF629}"+processedText
                    processedText="{Delay "+str(lastDelay)+"}"+"{Delay "+str(avgDelay)+"}"+processedText
                    if processedText[-1]!="\n":
                        processedText+="\n"
                    fixedEntry.append(processedText)
                    processedText,typingSound="",0
                    if addNewline:
                        fixedEntry.append("{Newline}\n")
                extraInst=getInstructions(currentLine,notSpreadTS)
                notSpreadTS=0
                processedInst+=fixInst(extraInst)
                currentLine=removeString(currentLine,extraInst)
            if currentLine=="\n":
                currentLine=""
    if processedInst:
        lastDelay=getLastDelay(processedInst)
        if processedInst[-1]!="\n":
            processedInst+="\n"
        fixedEntry.append(processedInst)
    if processedText:
        avgDelay=getAverageDelay(processedText,lastDelay)
        addNewline=checkIfTextEndsWNL(processedText)
        processedText=cleanInst(processedText)
        if typingSound==1:
            processedText="{0xF55C}{0xF63A}"+processedText
        elif typingSound==2:
            processedText="{0xF55C}{0xF629}"+processedText
        processedText="{Delay "+str(lastDelay)+"}"+"{Delay "+str(avgDelay)+"}"+processedText
        if processedText[-1]!="\n":
            processedText += "\n"
        fixedEntry.append(processedText)
        if addNewline:
            fixedEntry.append("{Newline}\n")
    return fixedEntry
def getInstructions(line,NTS=0):
    instLine,tmpLine,textFound="","",False
    while "{" in line:
        lparts=line.split("{",1)
        line=lparts[1]
        if lparts[0].strip():
            textFound=True
            break
        else:
            if not instLine:
                tmpLine+=lparts[0]
            else:
                instLine+=lparts[0]
            lparts=line.split("}",1)
            line=lparts[1]
            if instLine:
                instLine+="{"+lparts[0]+"}"
                if len(instLine)>23:
                    if instLine[:12]=="{Wait Press}"\
                            and instLine[22:]=="{Newline}"\
                            and instLine[13:18]=="Delay":
                        instLine="{Wait Press}{Pad00000}{Newline}"
                if len(instLine)>27:
                    if instLine[:12]=="{Wait Press}"\
                            and instLine[26:]=="{Newline}"\
                            and instLine[13:18]=="Delay":
                        instLine="{Wait Press}{Pad000000000}{Newline}"
                if len(instLine)>37:
                    if instLine[:12]=="{Wait Press}"\
                            and instLine[36:]=="{Newline}"\
                            and instLine[27:32]=="Delay":
                        instLine="{Wait Press}{Pad0000000000000000000}{Newline}"
            elif isHardInst(lparts[0]):
                if isTypingSound(lparts[0],line,NTS)==0:
                    instLine=tmpLine +"{"+lparts[0]+"}"
                    tmpLine=""
                else:
                    tmpLine+="{"+lparts[0]+"}"
            else:
                tmpLine+="{"+lparts[0]+"}"
    if not textFound:
        if not line.strip():
            if instLine:
                instLine+=line
                return instLine
            if len(tmpLine)>18:
                if tmpLine[-18:-13]=="Delay" \
                        and tmpLine[-9:]=="{Newline}":
                    tmpLine=tmpLine[:-18]+"Pad00000}{Newline}"
            tmpLine+=line
            return tmpLine
    return instLine
def fixInst(instLine):
    if "{Begin List}" in instLine:
        return "{Begin List}{Delay 00}"
    else:
        return instLine
def isHardInst(instruction):
    return ("List" in instruction or "Wait Press" in instruction or "0xF5" in instruction or "0xF6" in instruction)
def isTypingSound(instruction,line,TIMES=0):
    if instruction.count("F63A")>TIMES\
        or (instruction.count("F55C")>TIMES and line.count("F63A")>TIMES):
        return 1
    elif instruction.count("F629")>TIMES\
        or (instruction.count("F55C")>TIMES and line.count("F629")>TIMES):
        return 2
    return 0
def getLastDelay(line):
    delay=1
    if "{Close}" in line:
        line=line.split("{Close}")
        if len(line)>1:
            line=line[-1]
    while "{" in line:
        lparts=line.split("{",1)
        line=lparts[1]
        lparts=line.split("}",1)
        line=lparts[1]
        if isDelay(lparts[0]):
            delay=int(removeString(lparts[0],"Delay "),16)
    return delay
def getText(line,NTS=0):
    if getInstructions(line,NTS):
        return ""
    textLine=""
    while "{" in line:
        lparts=line.split("{",1)
        line=lparts[1]
        textLine+=lparts[0]
        lparts=line.split("}",1)
        line=lparts[1]
        if isHardInst(lparts[0]):
            if isTypingSound(lparts[0],line,NTS)==0:
                return textLine
        textLine+="{"+lparts[0]+"}"
    textLine+=line
    return textLine
def getAverageDelay(line,lastDelay):
    delay,delayTimes=lastDelay,1
    while "{" in line:
        lparts=line.split("{",1)
        line=lparts[1]
        lparts=line.split("}",1)
        line=lparts[1]
        if isDelay(lparts[0]):
            delay+=int(removeString(lparts[0],"Delay "),16)
            delayTimes+=1
    return math.floor(delay/delayTimes)
def isDelay(instruction):
    return True if "Delay" in instruction else False
def checkIfTextEndsWNL(text):
    if text:
        if text[-1]=="\n":
            text=text[:-1]
        if len(text)>8 and "{Newline}"==text[-9:]:
            return True
    return False
def removeString(line,subs):
    return line[len(subs) if len(subs)>0 else 0:]
def reapplyDelaysAndSound(delays,line):
    firstDelay=int((delays.split()[1]).split('}')[0])
    secondDelay=int((delays.split()[2]).split('}')[0])
    if "{0xF55C}{0xF63A}" in delays:
        auxLine,firstChar,stopAdding,secondDelay="",True,False,-2
        for char in line:
            if char=="{":
                stopAdding=True
            if (not stopAdding) and (char!=" ") and (char!="\n"):
                if firstChar:
                    auxLine+=char
                    firstChar=False
                else:
                    auxLine+="{0xF55C}{0xF63A}"+char
            elif (char!="\n"):
                auxLine+=char
            if char=="}":
                stopAdding=False
        line=auxLine+"{0xF55C}{0xF63A}"+"\n"
    elif "{0xF55C}{0xF629}" in delays:
        auxLine,stopAdding,secondDelay="",False,-9
        for char in line:
            if char=="{":
                stopAdding=True
            if (not stopAdding) and char!=" " and char!="\n":
                auxLine+="{0xF55C}{0xF629}"+char
            elif char!="\n":
                auxLine+=char
            if char=="}":
                stopAdding=False
        line=auxLine+"\n"
    if secondDelay<2 and firstDelay!=abs(secondDelay):
            return "{Delay 0"+str(abs(secondDelay))+"}"+line
    elif firstDelay>1:
        return "{Delay 01}"+line
    return line
def decompressText(data,i=0):
    if data.find(b"\xF0",i)+data.find(b"\xF1",i)+data.find(b"\xF2",i)+data.find(b"\xF3",i)==-4:
        return data
    ddata,idx=bytearray(),0
    while idx<len(data):
        if data[idx]&0xfc==0xf0:
            val=int.from_bytes(data[idx:idx+3],"big",signed=False)
            length=(val&0x030000)>>13+(val&0x00e000)>>17+4
            jump=val&0x001fff-math.floor((val&0x001fff)/256)*2
            if ddata:
                ddata+=data[idx-jump:idx-jump+length]
            else:
                ddata=data[:idx]+data[idx-jump:idx-jump+length]
            idx+=2
        elif ddata:
            ddata+=data[idx:idx+1]
        idx+=1
    return ddata
